print("import: 'geoana'")
import geoana

print("import: 'geoana.earthquake'")
import geoana.earthquake

print("import: 'geoana.em'")
import geoana.em

print("import: 'geoana.kernels._extensions'")
import geoana.kernels._extensions

